package org.processmining.ptlinearvisualization;

import java.awt.BorderLayout;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.swing.JComponent;
import javax.swing.JPanel;

import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.contexts.uitopia.annotations.Visualizer;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.linearlayouts.BlockConnector;
import org.processmining.linearlayouts.Connector;
import org.processmining.linearlayouts.LayoutPanel;
import org.processmining.linearlayouts.LinearLayout;
import org.processmining.linearlayouts.processtrees.ActionBlock;
import org.processmining.linearlayouts.processtrees.ChoiceBlock;
import org.processmining.linearlayouts.processtrees.LoopBlock;
import org.processmining.linearlayouts.processtrees.ParallelBlock;
import org.processmining.linearlayouts.processtrees.SequenceBlock;
import org.processmining.linearlayouts.processtrees.TwoLineActionBlock;
import org.processmining.processtree.Block;
import org.processmining.processtree.Block.And;
import org.processmining.processtree.Block.Xor;
import org.processmining.processtree.Block.XorLoop;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.Task;
import org.processmining.processtree.impl.AbstractBlock.Seq;

public class AgaProcessTreeVisualizer {

	@Plugin(name = "Folding Process Tree Visualizer",
			returnLabels = { "Visualizes process trees" },
			returnTypes = { JComponent.class },
			parameterLabels = { "Process tree" },
			userAccessible = true)
	@Visualizer
	@UITopiaVariant(
			affiliation = "Applied Geometric Algorithms, TU Eindhoven", 
			author = "Willem Sonke", 
			email = "w.m.sonke@tue.nl"
			)
	public static JComponent visualize(UIPluginContext context, ProcessTree tree) {
		Node root = tree.getRoot();
		final LinearLayout p = new LinearLayout();
		flatten(p, root, null);
		p.addMarginBlocks();
		
		final LayoutPanel panel = new LayoutPanel(p);
		
		/*JPanel buttonPanel = new JPanel(new FlowLayout());
		buttonPanel.add(new JButton("Export to Ipe") {{

			addActionListener(new ActionListener() {
				
				public void actionPerformed(ActionEvent e) {
					JFileChooser saveDialog = new JFileChooser();
					int result = saveDialog.showSaveDialog(panel);
					if (result != JFileChooser.APPROVE_OPTION) {
						return;
					}
					IpeGraphics2D.generateIpeOutput(p, saveDialog.getSelectedFile().getAbsolutePath());
				}
			});
		}});*/
		
		JPanel result = new JPanel(new BorderLayout());
		result.add(panel);
		//result.add(buttonPanel, BorderLayout.PAGE_END);

		return result;
	}
	
	private static LinearLayout flatten(LinearLayout p, Node root, Connector parentBackEdge) {

		if (root instanceof Task) {
			p.blocks.add(makeBlockFor(root));

		} else if (root instanceof Block) {
			Block block = (Block) root;

			if (block instanceof Seq) {
				for (Node child : block.getChildren()) {
					flatten(p, child, parentBackEdge);
				}

			} else if (block instanceof XorLoop) {
				BlockConnector e = new BlockConnector(p, p.connectors.size(), 0, 0, makeBlockFor(block), parentBackEdge);
				
				int firstBlock = p.blocks.size();
				flattenOnlySeq(p, block.getChildren().get(0), e);
				int lastBlock = p.blocks.size() - 1;

				e.from = lastBlock;
				e.to = firstBlock;
				
				if (e.to == e.from) {
					// self-loop; in that case make a LoopBlock (we don't support top-level self-loops)
					org.processmining.linearlayouts.Block loopBlock = makeBlockFor(root);
					assert(loopBlock instanceof LoopBlock);
					p.blocks.set(lastBlock, loopBlock);
					
				} else if (e.to > e.from) {
					// shouldn't happen!
					throw new RuntimeException("e.to <= e.from for " + e.to + " <= " + e.from);

				} else {
					if (block.getChildren().size() > 1) {
						if (!(block.getChildren().get(1) instanceof Task) ||
								!block.getChildren().get(1).getName().equals("tau")) {
							e.setBlock(makeBlockFor(block.getChildren().get(1)));
						}
					}
					p.connectors.add(e);
				}
				
			} else if (block instanceof Xor) {
				
				// is this a (pure) optional block?
				List<Node> children = new ArrayList<>(block.getChildren());
				boolean isPureOptional = false;
				/*if (children.size() == 2) {
					int firstBlock = p.blocks.size();
					if (children.get(0).getName().equals("tau")) {
						flattenOnlySeq(p, children.get(1), parentBackEdge);
						isPureOptional = true;
					} else if (children.get(1).getName().equals("tau")) {
						flattenOnlySeq(p, children.get(0), parentBackEdge);
						isPureOptional = true;
					}
					int lastBlock = p.blocks.size() - 1;
					if (isPureOptional) {
						if (firstBlock == lastBlock) {
							p.blocks.get(firstBlock).setOptional(true);
						} else {
							p.blocks.set(firstBlock, new OptionalLeftSideBlock(p.blocks.get(firstBlock)));
							p.blocks.set(lastBlock, new OptionalRightSideBlock(p.blocks.get(lastBlock)));
						}
					}
				}*/

				if (!isPureOptional) {
					p.blocks.add(makeBlockFor(block));
				}
				
			} else {
				p.blocks.add(makeBlockFor(block));
			}

		} else {
			throw new RuntimeException("Wait, no Task and no Block? Maybe I don't actually understand process trees :P");
		}
		
		return p;
	}

	private static LinearLayout flattenOnlySeq(LinearLayout p, Node root, Connector parentBackEdge) {

		if (root instanceof Task) {
			p.blocks.add(makeBlockFor(root));

		} else if (root instanceof Block) {
			Block block = (Block) root;

			if (block instanceof Seq) {
				for (Node child : block.getChildren()) {
					flattenOnlySeq(p, child, parentBackEdge);
				}

			} else if (block instanceof XorLoop) {
				BlockConnector e = new BlockConnector(p, p.connectors.size(), 0, 0, null, parentBackEdge);
				
				int firstBlock = p.blocks.size();
				flattenOnlySeq(p, block.getChildren().get(0), e);
				int lastBlock = p.blocks.size() - 1;

				e.from = lastBlock;
				e.to = firstBlock;
				
				if (e.to == e.from) {
					// self-loop; in that case make a LoopBlock (we don't support top-level self-loops)
					org.processmining.linearlayouts.Block loopBlock = makeBlockFor(root);
					assert(loopBlock instanceof LoopBlock);
					p.blocks.set(lastBlock, loopBlock);
					
				} else if (e.to > e.from) {
					// shouldn't happen!
					throw new RuntimeException("e.to <= e.from for " + e.to + " <= " + e.from);

				} else {
					if (block.getChildren().size() > 1) {
						if (!(block.getChildren().get(1) instanceof Task) ||
								!block.getChildren().get(1).getName().equals("tau")) {
							e.setBlock(makeBlockFor(block.getChildren().get(1)));
						}
					}
					p.connectors.add(e);
				}
				
			} else {
				p.blocks.add(makeBlockFor(block));
			}

		} else {
			throw new RuntimeException("Wait, no Task and no Block? Maybe I don't actually understand process trees :P");
		}
		
		return p;
	}
	
	private static org.processmining.linearlayouts.Block makeBlockFor(Node node) {

		if (node instanceof Task) {
			Task task = (Task) node;
			String name = task.getName();
			String[] parts = name.split("\\+");
			if (parts.length == 2) {
				return new TwoLineActionBlock(parts[0], "+" + parts[1]);
			}
			return new ActionBlock(task.getName());

		} else if (node instanceof Block) {
			Block block = (Block) node;

			if (block instanceof Seq) {
				List<org.processmining.linearlayouts.Block> children = new ArrayList<>(block.getChildren().size());
				for (Node child : block.getChildren()) {
					children.add(makeBlockFor(child));
				}
				return new SequenceBlock(children);

			} else if (block instanceof XorLoop) {
				// TODO handle redo block
				return new LoopBlock(makeBlockFor(block.getChildren().get(0)));

			} else if (block instanceof Xor) {
				List<Node> children = new ArrayList<>(block.getChildren());
				boolean isOptional = false;
				for (Iterator<Node> it = children.iterator(); it.hasNext();) {
					Node child = it.next();
					if (child instanceof Task) {
						Task task = (Task) child;
						if (task.getName().equals("tau")) {
							isOptional = true;
							it.remove();
						}
					}
				}

				if (children.size() == 1) {
					org.processmining.linearlayouts.Block b = makeBlockFor(children.get(0));
					b.setOptional(isOptional);
					return b;
				}

				List<org.processmining.linearlayouts.Block> childBlocks = new ArrayList<>(block.getChildren().size());
				for (Node child : children) {
					childBlocks.add(makeBlockFor(child));
				}
				org.processmining.linearlayouts.Block b = new ChoiceBlock(childBlocks);
				if (isOptional) {
					b.setOptional(true);
				}
				return b;

			} else if (block instanceof And) {
				List<org.processmining.linearlayouts.Block> children = new ArrayList<>(block.getChildren().size());
				for (Node child : block.getChildren()) {
					children.add(makeBlockFor(child));
				}
				return new ParallelBlock(children);
			}
		}
		
		return new ActionBlock("Unknown block type " + node.getClass().getSimpleName());
	}
}
